@@ -86,7 +86,7 @@ def needs_caller_reference(func):
 def api_action(*api):
 
     def decorator(func):
-        action = ''.join(api or map(str.capitalize, func.func_name.split('_')))
+        action = ''.join(api or map(str.capitalize, func.__name__.split('_')))
         response = ResponseFactory(action)
         if hasattr(boto.fps.response, action + 'Response'):
             response = getattr(boto.fps.response, action + 'Response')
